#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 21 10:16:58 2025

@author: pteysseyre
"""

''' Manages the file where the breakpoints and quiet amplitude and phase are written.'''

import os.path
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
import datetime as dt

def get_lastbreakpoint(la_station, la_date, path, quiet=False, check_yesterday=True):
    
    ''' Read the value of the last breakpoint, such as the time of detection, the slope, and the associated amplitude and phase.
    If quiet is True, it returns the values for the last quiet breakpoint.
    If there were no breakpoints today, return a series of zeros and create the file
    
    :param la_station: Call-sign of the station of interest
    :param la_date: Date of interest (e.g. '2023_11_05')
    :param path: path to where the results will be stored
    :param quiet: Boolean, is True if we want the last quiet breakpoint, False otherwise (default: False)
    :param check_yesterday: Boolean. If True, will get the last breakpoint of the previous day if needed (for file initialisation)
    
    :return: Last time, slope, amplitude, phase, quiet and DP of the last breakpoint. '''
    
    # Read the last breakpoint value OR create a new file if it is the first of the day
    # If there are no lines yet in the file, write header, using the last breakpoint of the previous day
    if not os.path.isfile(path+la_date+'_'+la_station+'.csv'):

        try:
            today = datetime.strptime(la_date, '%Y_%m_%d')

            if check_yesterday:
                try:
                    yesterday = today - timedelta(days=1)
                    yesterday = yesterday.strftime("%Y_%m_%d")
                    last_bp_t, last_bp_s, last_bp_a, last_bp_p, last_bp_q, last_bp_d = get_lastbreakpoint(la_station, yesterday, path, check_yesterday=False)
                    data = [[last_bp_t-24, last_bp_s, last_bp_a, last_bp_p, 1, 0]]

                    # Here, until branch 48-midnight is merged
                    data = [[0, 0, 0, 0, 1, 0]]

                except: # There were no data yesterday either
                    data = [[0, 0, 0, 0, 1, 0]]

            # NOTE: We force here the new breakpoint to be quiet. This is not strictly true, and may cause problems for flares
            # over midnight, but this is a quick fix for now.
            else: # If we don't check yesterday (to avoid infinite recursive loop)
                data = [[0, 0, 0, 0, 1, 0]]

        except: 
            data = [[0, 0, 0, 0, 1, 0]]

        df = pd.DataFrame(data, columns = ['Time (UT)', 'Slope (deg/hr)', 'Amplitude (pT)', 'Phase (deg)', 'Quiet', 'DP'])
        df.to_csv(path+la_date+'_'+la_station+'.csv', index=False)
        
    else :
        df = pd.read_csv(path+la_date+'_'+la_station+'.csv')
    
    if quiet:
        df = df.loc[df['Quiet']==1]
        
    last_bp = df.iloc[-1]
        
    return last_bp['Time (UT)'], last_bp['Slope (deg/hr)'], last_bp['Amplitude (pT)'], last_bp['Phase (deg)'], last_bp['Quiet'], last_bp['DP']

def write_newbreakpoint(la_station, la_date, path, time_now, p1, amp, phase, quiet, DP):
    
    ''' Update the state file by adding the new detected breakpoint.
    Note: because the function get_lastbreakpoint has been called earlier in the Realtimescript, there
    is no need to check that a file exist, or to create a new one
    
    :param la_station: Station call-sign
    :param la_date: Date of interest 
    :param path: path to where the results are stored
    :param time_now: time when the data was last read
    :param p1: Last breakpoint slope
    :param amp: Median amplitude over the last minute
    :param phase: Median phase over the last minute
    :param quiet: Boolean, is True if we are in quiet time, False otherwise
    :param DP: Phase increase compared to its quiet time value'''
    
    # Read the state file 
    df = pd.read_csv(path+la_date+'_'+la_station+'.csv')
    
    # Add the new values 
    df.loc[len(df)] = [time_now, p1, amp, phase, quiet, DP]
    
    # Write the new file
    df.to_csv(path+la_date+'_'+la_station+'.csv', index=False)
    
    return

def write_fluxestimate(best_guess, quiet, today, path, ending):

    ''' Writes in a file the best guess for the flux estimation and informations
    about how it was computed. This includes the DP values for each stations as
    well as the 'quiet' flag for each station.
    
    :param best_guess: Best estimate for the X-ray flux
    :param quiet: quiet value. It is proportional to the number of transmitters for which a quiet value was detected. If it is equal to 0, all transmitters have detected a flare time
    :param today: Datetime object, time in UT
    :param path: Destination of the file
    :param ending: Ending of the file name created'''

    time_now = today.hour + today.minute/60
    la_date = today.strftime("%Y_%m_%d")

    # Data to write
    data = [time_now, best_guess, quiet]

    # Read the last breakpoint value OR create a new file if it is the first of the day
    # If there are no lines yet in the file, write header, using the last breakpoint of the previous day
    if not os.path.isfile(path+la_date+'_' + ending+ '.csv'):

        df = pd.DataFrame([data], columns = ['Time (UT)', 'Best estimate', 'Quiet'])
        df.to_csv(path+la_date+'_' + ending+ '.csv', index=False)
        
    else :

        # Read the state file 
        df = pd.read_csv(path+la_date+'_' + ending+ '.csv')
        
        # Add the new values 
        df.loc[len(df)] = data
        
        # Write the new file
        df.to_csv(path+la_date+'_' + ending+ '.csv', index=False)
    
    return

def write_file_with_df(stations, path_to_result):

    '''Write a file with the current df values for each station
    
    :param stations: List of stations recorded
    :param path_to_result: Path to the file which will be written'''

    data = []
    names = []

    for st in stations:

        names.append(st.name)
        data.append(st.df)

    # Get date (for file name)
    today = dt.datetime.now(dt.timezone.utc) # The time zone is necessary to use get_sza later
    la_date = today.strftime("%Y_%m_%d")

    # Write file
    df = pd.DataFrame([data], columns = names)
    df.to_csv(path_to_result+la_date+'_df.csv', index=False)

    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  1 17:10:28 2025

@author: pteysseyre
"""

import pandas as pd
import datetime as dt
from datetime import timedelta, datetime
from pysolar.solar import get_altitude
import pytz
import numpy as np

""" This file deals with checking that we are in daytime.
It is based on previously computed sunrise and sunset times from https://nrc.canada.ca/en/research-development/products-services/software-applications/sun-calculator/

As a precaution, we define nighttime as between from one hour before sunset to one hour after sunrise.


History:
------------------

Updated in April 2025 to include calculation of the solar zenith angle

    
"""

def get_sza(today, lat, lon):
    ''' Returns the solar zenith angle (in rad) at the latitude/longitude given in input
    
    :param today: Datetime for the time of interest
    :param lat: Latitude of interest (in °)
    :param lon: Longitude of interest (in °)
    
    :return: sza, solar zenith angle (in rad)'''
    
    sza = float(90) - get_altitude(lat, lon, today)
    sza = sza*np.pi/180

    return sza

    
def get_sunrise_sunset(today, path):
    ''' Returns the sunrise (civil + 1 hr) and sunset (civil - 1 hr) times (in UT)
    
    :param today: datetime (may only contain the year, month and day)
    :param path: path to the precomputed sunrise/sunset times
    
    :return: sunrise and sunset times for this day (in UT)'''
    
    # Must be an easier way to do this, but get the year, month, and dayofyear of today.
    doy = today - dt.datetime(today.year, 1, 1, 0, 0, 0, 0, pytz.UTC)
    doy = doy.days
    
    # Get the sunrise / sunset time for that particular day
    file_sunrise = pd.read_csv(path, header = 1)
    
    sunrise = file_sunrise.iat[doy - 1, 3]
    sunrise = datetime.strptime(sunrise, '%H:%M') + timedelta(hours=1)
    sunrise = sunrise.time()
    
    sunset = file_sunrise.iat[doy - 1, 5]
    sunset = datetime.strptime(sunset, '%H:%M')  - timedelta(hours=1)
    sunset = sunset.time()
    
    return sunrise, sunset

    
def check_if_daytime(today, path):
    """ Check whether the current time is greater than the sunrise time and less than sunset
    for this particular station and day

    :param today: datetime (may only contain the year, month and day)
    :param path: path to the precomputed sunrise/sunset times 

    :return: Boolean, if we are in daytime
     """
    
    sunrise, sunset = get_sunrise_sunset(today, path)
  
    return (today.time() > sunrise and today.time() < sunset)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 21 11:47:33 2024

@author: pteysseyre
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 21 09:59:31 2024

@author: pteysseyre
"""

import numpy as np
from pymatreader import read_mat # To read the mau files
import os.path # Useful to determine if a file exist or not (to write headers in a new file)
import glob

#======================== Read mau files ======================================

def read_Narrowband_postprocessing(la_date, la_station, path, file_endings = ['_100A', '_100B', '_101A', '_101B']): 
    """
    Reads the data for the station and the day asked and stores them in the output arrays
    Needs the path to the data files
    Reads the .mat files (already ended)
    Written by C. Briand on 16//10/2022
    Converted to Python by P. Teysseyre on 18/03/2024
    
    :param la_date: Date when the files need to be read (e.g. '2023_01_25')
    :param la_station: Transmitter call sign
    :param path: path to the files
    :param file_endings: List of string. Endings of the files for amp_NS, phase_NS, amp_EW and phase_EW in that order"""

        
    Cal_NS = 0.14
    Cal_EW = 0.08
    
    nb_errors = 0
    
    # amp_NS
    list_of_files = glob.glob(path+'NC*'+la_station+'_100A.mat')
    latest_file = max(list_of_files, key=os.path.getctime)

    try :
        data = read_mat(latest_file)
        amp_NS = data['data']*Cal_NS
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - amp_NS')
        amp_NS = 0
        nb_errors += 1
        start_time = 0
    
    # phase_EW
    list_of_files = glob.glob(path+'NC*'+la_station+'_101B.mat')
    latest_file = max(list_of_files, key=os.path.getctime)

    try :
        data = read_mat(latest_file)
        phase_EW = data['data']
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - phase_EW')
        
        # If there is a pb with the phase_EW files, a solution is to take the phase_NS ones
        # phase_NS
        list_of_files = glob.glob(path+'NC*'+la_station+'_100B.mat')
        latest_file = max(list_of_files, key=os.path.getctime)

        try :
            data = read_mat(latest_file)
            phase_EW = data['data']
            start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
        except (TypeError, ValueError):
            print('TypeError - phase_NS')
            phase_EW = 0
            nb_errors += 1
            start_time = 0
    
    
    # amp_EW
    list_of_files = glob.glob(path+'NC*'+la_station+'_101A.mat')
    latest_file = max(list_of_files, key=os.path.getctime)

    try :
        data = read_mat(latest_file)
        amp_EW = data['data']*Cal_EW
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - amp_EW')
        amp_EW = 0
        nb_errors += 1
        start_time = 0

    # phase_NS
    list_of_files = glob.glob(path+'NC*'+la_station+'_100B.mat')
    latest_file = max(list_of_files, key=os.path.getctime)

    try :
        data = read_mat(latest_file)
        phase_NS = data['data']
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - phase_NS')
        phase_NS = 0
        nb_errors += 1
        start_time = 0
    
    Fs = data['Fs']
    tstep = 1/Fs/3600
    the_time = np.linspace(start_time, tstep*np.size(phase_EW), np.size(phase_EW))
    
    return (amp_NS, amp_EW, phase_NS, phase_EW, the_time)

def read_Narrowband_real_time(station, path, receiver): 

    """
    Reads the data for the station and the day asked and stores them in the output arrays
    Reads the .mat files still being written
    Needs the path to the files
    Written by C. Briand on 16//10/2022
    Converted to Python on 18/03/2024
    
    :param station: 'station' class instance 
    :paraam path: Path to the files being read
    :param receiver: Receiver class instance"""

    nb_errors = 0
    
    # amp_NS
    list_of_files = glob.glob(path+'*'+station.name+receiver.file_endings[0] + '.mau')
    latest_file = max(list_of_files, key=os.path.getctime)
    #fname = shutil.copy2(f, pathdest) # If we want to copy the files, decomment the line and change the following f to fname
    try :
        data = read_mat(latest_file)
        amp_NS = data['data']*station.Cal_NS
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - amp_NS')
        amp_NS = 0
        nb_errors += 1
        start_time = 0
    #os.remove(fname)
    
    # phase_EW
    list_of_files = glob.glob(path+'*'+station.name+receiver.file_endings[3] + '.mau')
    latest_file = max(list_of_files, key=os.path.getctime)
        #fname = shutil.copy2(f, pathdest)
    try :
        data = read_mat(latest_file)
        phase_EW = data['data']
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - phase_EW')
        phase_EW = 0
        nb_errors += 1
        start_time = 0
    
    
    # amp_EW
    list_of_files = glob.glob(path+'*'+station.name+receiver.file_endings[2] + '.mau')
    latest_file = max(list_of_files, key=os.path.getctime)
        #fname = shutil.copy2(f, pathdest)
    try :
        data = read_mat(latest_file)
        amp_EW = data['data']*station.Cal_EW
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - amp_EW')
        amp_EW = 0
        nb_errors += 1
        start_time = 0
        #os.remove(fname)

    # phase_NS
    list_of_files = glob.glob(path+'*'+station.name+receiver.file_endings[1] + '.mau')
    latest_file = max(list_of_files, key=os.path.getctime)
        #fname = shutil.copy2(f, pathdest)
    try :
        data = read_mat(latest_file)
        phase_NS = data['data']
        start_time = data['start_hour'] + data['start_minute']/60 + data['start_second']/3600
    except (TypeError, ValueError):
        print('TypeError - phase_NS')
        phase_NS = 0
        nb_errors += 1
        start_time = 0
        #os.remove(fname)
    
    Fs = data['Fs']
    tstep = 1/Fs/3600
    the_time = np.linspace(start_time, tstep*np.size(phase_EW), np.size(phase_EW))
    
    return (amp_NS, amp_EW, phase_NS, phase_EW, the_time, nb_errors)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug 27 11:42:43 2024

@author: pteysseyre
"""

import numpy as np


def find_break_point_realtime(the_time, data, delta = 0.1):
    """
    Based on the incremental algorithm presented by Guralnik & Srivastava (1999)

    :param the_time: time-array
    :param data: the phase data that has just been read (time-averaged & smoothed if needed)
    :param delta: in percent/100 (default : 0.1). This is the threshold over which we 
     decide that a breakpoint is indeed a breakpoint and not due to noise. In
     Guralnik & Srivastava's paper, this is also called delta and only appears 
     for the incremental algorithm.

    :return: - detec: Boolean, 'True' if a breakpoint was detected, 'False' if not
     - p1: Slope of the detected breakpoint

    """
    
    L = len(data)    
    
    if L >  9: # At least 9 points are necessary for this
        
        # Compute likelihood assuming no breakpoints:
        likelihood_without_bp, n = _compute_likelihood_criteria(0, L, the_time, data)
        
        # Compute the likelihood criteria, assuming a breakpoint 5 points ago:
        likelihood_with_bp, p1 = _compute_likelihood_criteria_several_change_points(the_time, data, np.array([0, L-6, L-1]))
        
        # Determine whether there is a new breakpoint
        detec = (likelihood_without_bp - likelihood_with_bp)/likelihood_without_bp > delta
    
    else:
        
        detec = False
        p1 = 0
    
    return detec, p1
            
    
    
def detect_change_points(the_time, data, threshold):
    """Based on Guralnik, V. & Srivastava, J. (1999)
    Detect change_points in the 'data' array in post_processing (bash algorithm)
     
    :param the_time: time-array
    :param data: Data array
    :param threshold: Criterion used to end the search for breakpoints. (s in the article)

    :return: change_points: time of breakpoints found in the data array """
    
    # Initialisation
    change_points = np.array([0, len(data)])
    likelihood, dummy = _compute_likelihood_criteria(change_points[0], change_points[1], the_time, data)
    last_likelihood = likelihood
    stopping_criterion = 1 
    loop_counter = 0
    
    while stopping_criterion > threshold and loop_counter < 1e3:
        
        loop_counter += 1
        N = len(change_points) - 1 # Number of segments
        
        
        for s in range(0, N):
            
            if change_points[s + 1] - change_points[s] > 5:
                new_candidate = _find_candidate_detection(change_points[s], change_points[s + 1], the_time, data)
                temp_change_points = np.insert(change_points, s + 1, new_candidate)
                new_likelihood_criterion = _compute_likelihood_criteria_several_change_points(the_time, data, temp_change_points)
                
                if new_likelihood_criterion < likelihood:
                    likelihood = new_likelihood_criterion
                    new_change_points = temp_change_points
                
        # Prepare for next iteration
        change_points = new_change_points
        stopping_criterion = (last_likelihood - likelihood)/last_likelihood
    
    return change_points
        
    

def _compute_likelihood_criteria_several_change_points(the_time, data, change_points):
    """ Based on Guralnik, V., & Srivastava, J. (1999). Computes the total likelihood criterion for data, 
     knowing that there are change_points at specified locations
    
    :param the_time: time-array
    :param data: Data in which to look for change points
    :param change_points: list of indices of the changepoints already computed.
     The first index is 0 (for the beggining of the 'data' array), the last is
     the number of elements in the 'data' array.
     
    :return: - total_L: total likelihood criterion
    - a: slope associated with the last breakpoint"""
    
    N = len(change_points)-1
    total_L = 0
    
    for k in range(0, N):
        likelihood_criteria, a = _compute_likelihood_criteria(change_points[k], change_points[k+1], the_time, data)
        total_L += likelihood_criteria
    
    return total_L, a

def _find_candidate_detection(i, j, the_time, data):
    """Based on Guralnik, V., & Srivastava, J. (1999).,
    
    Finds the best candidate to be a change-point in the array data[i:j]. 
    
    :param i: indice of the start of the interval of interest
    :param j: indice of the end of the interval
    :param the_time: time-array
    :param data: data array
    
    :return: the indice in the array data (! not data[i:j]!) of this best candidate.
    """
    
    optimal_likelihood_criteria = 1e14 # Minimum likelihood criteria. The point with the lowest likelihood criteria will be the best candidate
    split = 0
    
    for k in range(i+2, j-2):
        
        likelihood_first_segment, dummy = _compute_likelihood_criteria(i, k, the_time, data)
        likelihood_second_segment, dummy = _compute_likelihood_criteria(k+1, j, the_time, data)
        likelihood_all_segments = likelihood_first_segment + likelihood_second_segment
        
        if likelihood_all_segments < optimal_likelihood_criteria:
            optimal_likelihood_criteria = likelihood_all_segments
            split = k
    
    return split

def _compute_likelihood_criteria(i, j, the_time, data):
    """ Computes the likelihood criteria assuming an heteroscedastic error model. 
    Based on Guralnik, V., & Srivastava, J. (1999, August). Event detection from time series data. In Proceedings of the fifth ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 33-42).
    
    :param i: indice of the start of the interval of interest
    :param j: indice of the end of the interval
    :param the_time: time-array
    :param data: data array"""
    
    m = j - i + 1 #Number of elements in the the_time & data arrays
    
    a, b = np.polyfit(the_time[i:j+1], data[i:j+1], 1) # Use a linear model to fit the segment. THIS CAN BE CHANGED !
    
    rss = np.sum(np.square(a*the_time[i:j+1] + b - data[i:j+1]))
    
    L = m * np.log(rss/m)
    
    return L, a
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 15 14:48:32 2025

@author: pteysseyre
"""

import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.basemap import Basemap
import matplotlib.cm as cm
from operator import attrgetter

def plot_map(stations_l, thisreceiver, today, path_to_results, Nemax=1e10):
    '''Plots the current state on a geographical map
    
    :param stations_l: list of station class instances
    :param this_receiver: Receiver class instance
    :param today: Datetime (in UT !) of interest
    :param path_to_results: path to where the plot will be stored
    :param Ne_max: Maximum possible electron density to ajust the colorbar (default, 1e11 m^-3). Automatically adjusted if Ne for a function is more than Nemax.'''

    # Automatically center the map on the receiver/transmitters' paths
    max_lat = max(stations_l, key=attrgetter('lat'))
    max_lat = max_lat.lat
    max_lat = max(max_lat, thisreceiver.lat) + 5
    max_lon = max(stations_l, key=attrgetter('lon'))
    max_lon = max_lon.lon
    max_lon = max(max_lon, thisreceiver.lon) + 5
    min_lat = min(stations_l, key=attrgetter('lat'))
    min_lat = min_lat.lat
    min_lat = min(min_lat, thisreceiver.lat) - 5
    min_lon = min(stations_l, key=attrgetter('lon'))
    min_lon = min_lon.lon
    min_lon = min(min_lon, thisreceiver.lon) - 5
    max_Ne = max(stations_l, key=attrgetter('Ne'))
    max_Ne = max_Ne.Ne

    if max_Ne > Nemax:
        Ne_max = max_Ne

    fig, ax = plt.subplots(1, 1, figsize=(6, 8))  # setup the plot
    m = Basemap(llcrnrlon=min_lon,
            llcrnrlat=min_lat,
            urcrnrlon=max_lon,
            urcrnrlat=max_lat,
            projection='merc')
    m.drawmapboundary(fill_color='white', linewidth=0)
    m.fillcontinents(color='grey', alpha=0.7, lake_color='grey')
    m.drawcoastlines(linewidth=0.1, color="white")
    
    norm = mpl.colors.LogNorm(vmin=1e8, vmax=Nemax)
    cmap = cm.RdYlGn_r
    sm = cm.ScalarMappable(norm=norm, cmap=cmap)
    m.nightshade(today, alpha=0.25)
        

    for st in stations_l:

        color  = sm.to_rgba(st.Ne)

        if st.flag == 0:
            m.drawgreatcircle(st.lon, st.lat, thisreceiver.lon, thisreceiver.lat, linestyle = '-',  linewidth=3, color=color)
            
            if st.name == 'GQD':
                plt.annotate(st.name, xy=m(st.lon+3, st.lat), xytext =(-40, -4), textcoords='offset points', fontsize = 12)
            else:
                plt.annotate(st.name, xy=m(st.lon+3, st.lat), xytext =(1, -5), textcoords='offset points', fontsize = 12)

    
    
    sm = cm.ScalarMappable(norm=norm, cmap=cmap)
    sm.set_array([]) # can be an empty list, only needed for matplotlib < 3.1
    # ...
    cb = fig.colorbar(sm, ax=ax, orientation = 'horizontal', pad = 0.02)
    cb.set_label('Ne ($m^{-3}$)', rotation=360)
    plt.title(today.strftime("%d/%m/%y %H:%M:%S"))
    plt.savefig(path_to_results + 'last_map.pdf')
    plt.close()

    return 





import os
import vlf4ions.class_definition as cd
import numpy as np

dirname = os.path.dirname(__file__)

def test_get_pdf_coeff():

    errors = []
    NRK = cd.station('NRK', 63.85, -22.47, 37.5, 22.02, ':15', df=1.4136e-4)
    path_to_files = os.path.join(dirname, '')
    
    NRK.DP = 0
    NRK.get_pdf_coeff(path_to_files)
    if NRK.mu != -6 or NRK.sigma != 1:
        errors.append('Problem with the prior')
    
    NRK.DP = 10
    NRK.get_pdf_coeff(path_to_files)
    if np.abs(NRK.mu + 4.66905892623818) > 1e-4 or np.abs(NRK.sigma - 0.199174407925015) > 1e-4:
        errors.append('Problem with the first row')

    NRK.DP = 200
    NRK.get_pdf_coeff(path_to_files)
    if np.abs(NRK.mu + 4.10760899331066) > 1e-4 or np.abs(NRK.sigma - 0.481437593395161) > 1e-4:
        errors.append('Problem with the last row')
    
    # assert no error message has been registered, else print messages
    assert not errors, "errors occured:\n{}".format("\n".join(errors))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 21 10:44:38 2024

@author: pteysseyre
"""

import numpy as np
import vlf4ions.sunrise_sunset as srst

def flags(amp_NS, amp_EW, previous_nb_points, today, path, this_receiver):
    """Flags 1 if the antenna is down, 2 if the transmitter is down, 3 if it is nighttime
    
    :param amp_NS: Amplitude in the N/S direction
    :param amp_EW: Amplitude in the E/W direction
    :param previous_nb_points: Number of points read last time
    :param today: Datetime, instant of file reading
    :param path: path to the precomputed sunrise/sunset times
    :param this_receiver: receiver class instance
    
    :return: - flag: flag indicating the state of the transmitter/receiver
        - length_data: number of points read this time
        - data_last_min: median total amplitude over the last minute"""
    
    flag = 0
    data_last_min = 0
    length_data = min(len(amp_NS), len(amp_EW)) # this is necessary as there may be one more point in amp_EW than amp_NS, as it is read later
    
    #Check that we are in daytime
    daytime = srst.check_if_daytime(today, path)
    
    if daytime:
        # Antenna down: the number of data points in the last written file stays the same
        if length_data == previous_nb_points:
            flag = 1
        else:
            previous_nb_points = length_data
                    
            if length_data > 60:
                amp_NS = amp_NS[-60:]
                amp_EW = amp_EW[-60:]
            else:
                amp_NS = amp_NS[-length_data:]
                amp_EW = amp_EW[-length_data:]
                
            amp = np.sqrt(np.square(amp_NS) + np.square(amp_EW))
            data_last_min = np.median(amp)
        
        if data_last_min < this_receiver.threshold and flag != 1: # If the amplitude is too low, the station is not transmitting
            flag = 2
    else:
        flag = 3
        
    return flag, length_data, data_last_min
    


import vlf4ions.manage_statefile as ms
import datetime as dt
from vlf4ions.class_definition import station
from datetime import datetime, timedelta
import os
import pandas as pd
import pytz
import glob

dirname = os.path.dirname(__file__)


def test_write_fluxestimation():

    today = dt.datetime(2000, 1, 1, 12, 3, 0, 0, pytz.UTC)
    path_to_files = os.path.join(dirname, '')
    filename = path_to_files + '2000_01_01_test.csv'

    # Remove file if it exists
    if os.path.exists(filename):
        os.remove(filename)

    # Test file creation 
    ms.write_fluxestimate(1e-5, 0, today, path_to_files, 'test')

    # Test adding new data
    today = dt.datetime(2000, 1, 1, 12, 5, 0, 0, pytz.UTC)
    ms.write_fluxestimate(1e-6, 1, today, path_to_files, 'test')

    # Read the file
    df = pd.read_csv(path_to_files + '2000_01_01_test.csv')
    first_line = df.iloc[0]
    second_line = df.iloc[1]


    assert(df.size == 6 and first_line['Time (UT)'] == 12.05 and second_line['Quiet'] == 1)

def test_write_df():

    NRK = station('NRK', 56.2, -7.6, 37.5, 22.02, ':10', df=1.4136e-4, orientation=0)
    NAA = station('NAA', 51.6, -33.5, 24.0, 32.39, ':18', df=3.1128e-5, orientation=0)
    stations = [NRK, NAA]

    path_to_results = os.path.join(dirname, 'Results/')
    
    # Remove all previous files
    filename = glob.glob(path_to_results + '*_df.csv')
    filename = max(filename, key=os.path.getctime) # To get a filename and not a list

    # Remove file if it exists
    if os.path.exists(filename):
        os.remove(filename)

    ms.write_file_with_df(stations, path_to_results)

    # Read the file for TODAY (list of files needs to be reload)
    filename = glob.glob(path_to_results + '*_df.csv')
    filename = max(filename, key=os.path.getctime) # To get a filename and not a list
    df = pd.read_csv(filename)
    first_line = df.iloc[0]

    assert(df.size == 2 and first_line['NAA'] == 3.1128e-5)

def test_manage_bp():

    path = os.path.join(dirname, 'Results/')

    # Remove all previous files
    filename = glob.glob(path + '*Test_ms.csv')

    # Remove file if it exists
    for f in filename:
        if os.path.exists(f):
            os.remove(f)

    # Create parameters for test
    la_station = 'Test_ms'

    # Get today's date in a good format
    today = dt.datetime.now(dt.timezone.utc) # The time zone is necessary to use get_sza later
    la_date = today.strftime("%Y_%m_%d")

    # Quiet is False
    the_time, slope, amp, phase, quiet, DP = ms.get_lastbreakpoint(la_station, la_date, path)

    # Write new breakpoint
    ms.write_newbreakpoint(la_station, la_date, path, 2, 2, 2, 2, 0, 2)

    # Get last breakpoint
    the_time_2, slope_2, amp_2, phase_2, quiet_2, DP_2 = ms.get_lastbreakpoint(la_station, la_date, path)

    # Get last quiet breakpoint
    the_time_1, slope_1, amp_1, phase_1, quiet_1, DP_1 = ms.get_lastbreakpoint(la_station, la_date, path, quiet=True)

    assert((the_time == 0) and quiet == 1 and quiet_1 == 1 and quiet_2 == 0 and slope_2 == 2 and amp_2 == 2 and phase_1 == 0 and DP == 0)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 21 10:00:48 2024

@author: pteysseyre
"""

from setuptools import find_packages, setup

setup(
    name='vlf4ions',
    packages=find_packages(include=['vlf4ions']),
    version='1.3.3-beta',
    description='Functions necessary to analyse in real time the output of an AWESOME antenna and detect any solar flare',
    author='P. Teysseyre',
    install_requires=['pymatreader==1.0.0', 
                      'schedule==1.2.2', 
                      'pandas==2.2.3', 
                      'numpy==1.26.4', 
                      'basemap==1.4.1', 
                      'matplotlib==3.8.0',
                      'pysolar==0.13',
                      'pytz==2025.2',
                      'scipy==1.15.2',
                      'datetime==5.5', 
                      ],
)


import vlf4ions.class_definition as cd
import vlf4ions.manage_statefile as ms
import numpy as np
import os
import pandas as pd
import datetime as dt
from datetime import datetime

dirname = os.path.dirname(__file__)

def test_email_alert_update_body():
    test_alert = cd.email_alerts('test_subject', 'test_body', 'test_sender', 'test_password', ['test_recipients'])
    test_alert.update_detection_body('NRK', 0)
    #print(test_alert.body)
    assert((test_alert.body != 'test_body'))


# TODO: test send_email

def test_sliding_median():
    test_array = np.arange(6)
    res = cd.sliding_median(test_array, 3)

    assert(all(res == [1., 2., 3., 4.]))


def test_two_hours_data():

    # Station 
    st = cd.station('Test', 0, 0, 0, 0.01, ":00")
    st.p1 = 0
    Nancay = cd.receiver('NC', lat = 47.3755933, lon = 2.1944333, threshold=-1)

    # Paths
    path_to_data = os.path.join(dirname, 'Data/')

    # Create breakpoint file (must have the date in the name, that's why we create 
    # it each time we test the function)
    today = dt.datetime.now(dt.timezone.utc) # The time zone is necessary to use get_sza later
    la_date = today.strftime("%Y_%m_%d")
    ms.get_lastbreakpoint('Test', la_date, path_to_data, check_yesterday=False)

    # Update breakpoint with synthetic data longer than two hours
    st.update_breakpoints('', path_to_data, path_to_data, [], [], Nancay)

    # Remove created file
    filename = path_to_data + la_date + '_Test.csv'

    # Remove file if it exists
    if os.path.exists(filename):
        os.remove(filename)

    # The slope in the test file is exactly 1 for the past 9 minutes and 0 before
    # Since the Test station has been set with a threshold of 0.01, if the code 
    # works, the detected slope should be 1 (because the data is longer than two 
    # hours without a breakpoint). It thus should counts as a flare detection, 
    # except we actually set quiet to be 1 manually.
    assert(st.p1 == 1 and st.quiet == 1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 22 14:10:40 2025

@author: pteysseyre
"""

import vlf4ions.read_files as rf
import vlf4ions.detect_change_points as dcp
import vlf4ions.correct_phase as correct_phase
import vlf4ions.flag as fl
import vlf4ions.analyse_change_points as acp
import vlf4ions.manage_statefile as msf

import numpy as np
import datetime as dt # Necessary to use timezone
from datetime import datetime, timedelta
import logging

import pandas as pd

# For alerts
import smtplib
from pathlib import Path
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
from email import encoders


#----------------------- Receiver description -----------------
class receiver:
    ''' Characteristics of the receiver (for plotting purposes mainly and for sending alerts) 
    
    :param name: Name of the receiver. Must match the name in the VLF data file names
    :param lat: Latitude of the receiver (for plotting)
    :param lon: Longitude of the receiver (for plotting)
    :param threshold: Amplitude threshold bellow which we declare a station not transmitting (default: 0.5)
    :param nb_errors: Number of reading errors, we keep track of them to alert in case of malfunction'''

    def __init__(self, name, lat, lon, threshold=0.5, nb_error=0, file_endings = ['_100A', '_100B', '_101A', '_101B']):
        self.name = name
        self.lat = lat
        self.lon = lon
        self.threshold = threshold
        self.nb_error = nb_error
        self.file_endings = file_endings

#---------------------- Alerts, warnings and such --------------
class email_alerts: 
    """ Alerts sent in case of flare detection, antenna problems or such
     
    :param subject: Subject of the email alert
    :param body: Body of the email alert
    :param sender: Email adress of the sender
    :param password: Password of the email address of the sender
    :param recipiends: Recipients of the email alert (must be a list of strings)
    :param threshold: Flux threshold. If the probability of the flux being above this threshold 
        is 10% or more, the alert will be sent (default: 0)
    :param files: List of files to include as attachment (default: empty)
    :param send_last_time: Computed internally; used to keep track of alerts already sent to avoid spamming people
    :param send_each_change: Boolean. If true, the recipients will receive an alert each time the flux estimation changes, as long 
        as there is at least a 10% probability of the flux being above the threshold  """
    
    def __init__(self, subject, body, sender, password, recipients, threshold=0, files=[], sent_last_time=0, send_each_change=False):
        self.subject = subject
        self.body = body
        self.sender = sender
        self.password = password
        self.recipients = recipients
        self.threshold = threshold # Only send the alerts if the proba of the flare is 10% above a threshold
        self.files = files # Files to attach to the email
        self.sent_last_time = sent_last_time # Boolean, only useful for alerts on probas
        self.send_each_change = send_each_change # Boolean, if True send each time the proba change
    
    def update_detection_body(self, callsign, p1):

        """ Just updates the body of the email with the relevant informations in case of flare detection"""

        self.body = 'Detection with ' + callsign + '\n Slope : ' + str(p1) + ' deg/hr.'

    def send_email(self): 

        """ Sends the email from the sender email adress to the recipients"""
        
        msg = MIMEMultipart()
        msg['From'] = self.sender
        msg['To'] = COMMASPACE.join(self.recipients)
        msg['Date'] = formatdate(localtime=True)
        msg['Subject'] = self.subject

        msg.attach(MIMEText(self.body))

        for path in self.files:
            part = MIMEBase('application', "octet-stream")
            with open(path, 'rb') as file:
                part.set_payload(file.read())
            encoders.encode_base64(part)
            part.add_header('Content-Disposition',
                            'attachment; filename={}'.format(Path(path).name))
            msg.attach(part)
        
        with smtplib.SMTP_SSL('smtp.gmail.com', 465) as smtp_server:
           smtp_server.login(self.sender, self.password)
           smtp_server.sendmail(self.sender, self.recipients, msg.as_string())  
           smtp_server.quit()
        
        self.sent_last_time = 1

#------------------- Stations and related functions --------------

def sliding_median(arr, window):
    return np.median(np.lib.stride_tricks.sliding_window_view(arr, (window,)), axis=1)
   
class station:
    ''' Characteristics of the transmitter.
    
    :param name: call-sign (e.g. 'GVT')
    :param lat: latitude of the transmitter
    :param lon: latitude of the transmitter
    :param freq: Frequency (kHz)
    :param detection_threshold: Phase slope above which we detect a solar flare
    :param reading_time: second after the minute at which we read the data. NOTE: They should be unique for each station
    :param time_average: Time resolution of the smoothed signal (default: 60 s)
    :param delta: Delta parameter of the incremental algorithm (Guralnik & Srivastava, 1999. Default:0.1)
    :param nb_data_points: number of data points last time the file was read (default : 0, used for flagging)
    :param Ne: Average electron density computed in the path (not done yet, but I needed this for data analysis)
    :param DA/DP: Computed internally, variations of the amplitude and phase compared to quiet times
    :param mu/sigma: This is computed internally, and is the average and std of the flux probability density function given DA and DP
    :param orientation: Is 1 if the phase considered is E/W, 0 if it is N/S (default: 1). The idea is that the phase can be much noisier in 
        one direction, so the optimal orientation is left as a choice for the user
    
    ---------------
    History: 
        Written by P. Teysseyre in January 2025
        Re-written under current format in April 2025
    
    
    
    '''
    def __init__(self, name, lat, lon, freq, detection_threshold, reading_time, Cal_NS = 0.14, Cal_EW = 0.08, df = 0, time_average = 60, delta = 0.1, nb_data_points = 0, nb_error = 0, Ne=1e8, DP=0, DA=0, p1=0, quiet=1, flag=0, mu=-6, sigma = 1, orientation=1):
        self.name = name # Call-sign
        self.lat = lat # Latitude
        self.lon = lon # Longitude
        self.freq = freq # Frequency
        self.detection_threshold = detection_threshold 
        self.reading_time = reading_time # When the data is read (must not be the same for two stations)
        self.Cal_NS = Cal_NS # Calibration number in the N/S orientation
        self.Cal_EW = Cal_EW # Calibration number in the E/W orientation
        self.time_average = time_average # Time-resolution
        self.delta = delta # Delta parameters for the detection (see Guralnik & Srivasta, 1999)
        self.nb_data_points = nb_data_points 
        self.Ne = Ne
        self.df = df # For phase correction
        self.DP = DP # Phase increase compared to quiet
        self.DA = DA # Amplitude increase compared to quiet
        self.p1 = p1 # Last breakpoint slope
        self.quiet = quiet # Is 1 if it is quiet
        self.flag = flag # Is 1 if the station is transmitting
        self.mu = mu # Average of the flux pdf
        self.sigma = sigma # Std of the flux pdf
        self.orientation = orientation # 1 to look for the E/W phase, 0 for the N/S (default 1)

    def update_df(self, this_receiver, path_to_data, epsilon=5, nb_days=30):

        '''Updates the value of df for the station, as it depends on the transmitter and may vary in time
        
        :param receiver: Receiver class instance
        :param path_to_data: Path to the data files
        :param epsilon: Maximum phase jump (in °) that we allow on average (default: 5)
        :param nb_days: Number of days on which we want to compute df (default: 30)'''

        # Get today's date, date_inf (yesterday - nb_days) and date_sup (yesterday)
        today = dt.datetime.now(dt.timezone.utc) # The time zone is necessary to use get_sza later
        date_sup = today - dt.timedelta(days=1)
        date_inf = date_sup - dt.timedelta(days=nb_days)

        # Compute the new df
        df = correct_phase.compute_df(date_inf, date_sup, self, path_to_data, this_receiver.file_endings, epsilon)

        # Update the df value
        self.df = df
    
    def update_breakpoints(self, path_breakpoint, path_mau, path_sunrise, alerts_antenna, alerts_detection, this_receiver):
        """ Checks the phase and detects perturbations due to flares. If needed, send alerts to users
        
        :param path_breakpoint: path to the files where the breakpoints are stores
        :param path_mau: path to the data being written
        :param path_sunrise: path to the files with precomputed sunrise and sunset times
        :param alerts_antenna: instance of 'alert' class, alert to send if the antenna is down
        :param alerts_detection: instance of 'alert' class, alert to send if there is a detection
        :param this_receiver: Receiver class instance
        """

        try:
            # Get today's date in a good format
            today = dt.datetime.now(dt.timezone.utc) # The time zone is necessary to use get_sza later
            la_date = today.strftime("%Y_%m_%d")
            
            # Read data & corrects the phase
            amp_NS, amp_EW, phase_NS, phase_EW, the_time, nb_error = rf.read_Narrowband_real_time(self, path_mau, this_receiver)

            if self.orientation==1:
                phase = phase_EW
            else:
                phase = phase_NS
            phase, jumps = correct_phase.correctionphaseparDw(self, phase, the_time)
            
            # Flag the data if needed
            # CAREFUL: amp is the median of the amplitude recorded over the last minute at most
            path_sunrise = path_sunrise +self.name+'_sunrise_sunset.txt'
            flag, nb_data_points, amp = fl.flags(amp_NS, amp_EW, self.nb_data_points, today, path_sunrise, this_receiver)
            self.nb_data_points = nb_data_points
            self.flag = flag
            
            if flag == 1: # The antenna has stopped
                self.nb_error += 1
                
                
            elif flag == 0:
                
                this_receiver.nb_error = 0
                
                # Time average the data & apply median filter
                time_average = np.min([self.time_average, len(the_time)])
                the_time = sliding_median(the_time, self.time_average)
                the_time = the_time[0::time_average]
                phase = sliding_median(phase, self.time_average)
                phase = phase[0::time_average]
                
                # Read the last breakpoint value OR create a new file if it is the first of the day
                last_breakpoint, last_p1, last_amp, last_phase, quiet, DP = msf.get_lastbreakpoint(self.name, la_date, path_breakpoint)
                
                # Keep the data after last breakpoint
                mask_time = (the_time > last_breakpoint)
                the_time = the_time[mask_time]
                phase = phase[mask_time]
                
                # Shorten the data if it is too long 
                # Also consider that it is quiet time, as there would have been more changes if it was flare time
                if the_time[-1] - the_time[0] > 2:
                    two_hours = int(np.floor(7200/self.time_average)) # Number of indices in those two hours
                    phase = phase[-1 - two_hours:]
                    the_time = the_time[-1 - two_hours:]

                    # Look if there are new breakpoints now
                    detec, p1 = dcp.find_break_point_realtime(the_time, phase, self.delta)

                    # NOTE: Here, the p1 value is the one corresponding to the last nine minutes of data

                    # Write new breakpoint in file
                    msf.write_newbreakpoint(self.name, la_date, path_breakpoint, the_time[-1], p1, amp, phase[-1], 1, 0)
                
                    print('New breakpoint, after two quiet hours')

                    # Change the station's characteritics
                    self.quiet = 1
                    self.DA = 0
                    self.DP = 0
                    self.p1 = p1
            
                else:
                    # Look if there are new breakpoints now
                    detec, p1 = dcp.find_break_point_realtime(the_time, phase, self.delta)
                
                    # If there is a new breakpoint, write it in a file
                    if detec:   
                    
                        # Get last QUIET breakpoint
                        quiet_breakpoint, quiet_p1, quiet_amp, quiet_phase, quiet_quiet, quiet_DP = msf.get_lastbreakpoint(self.name, la_date, path_breakpoint, quiet=True)
                        # NOTE: quiet_quiet is useless and is always 1 (same for quiet_DP who is always 0)
                        
                        quiet_now, prev_timedecay, DP, DA = acp.analyse_breakpoint(the_time[-1], p1, amp, phase, the_time, quiet, quiet_phase, quiet_amp, last_breakpoint, last_p1, self.detection_threshold, quiet_p1, quiet_breakpoint)
                        
                        if self.quiet == 1 and p1 > 1000:
                            print('Breakpoint due to bad correction of phase')
                            self.flag = 5
                            quiet_now = 1
                            DP = 0
                            DA = 0
                        elif quiet_now == 1:
                            DP = 0
                            DA = 0
                            
                        
                        # Write new breakpoint in file
                        msf.write_newbreakpoint(self.name, la_date, path_breakpoint, the_time[-1], p1, amp, phase[-1], quiet_now, DP)
                    
                        print('New breakpoint !')

                        # Update the station
                        self.p1 = p1
                        self.quiet = quiet_now
                        self.DP = DP
                        self.DA = DA
                        
                        # Send alerts if needed
                        if quiet_now == 0 and quiet == 1:
                            
                            alerts_detection.update_detection_body(self.name, p1)
                            alerts_detection.send_email()
                    
                    print('Done - ', today.strftime("%H:%M:%S"), '-', self.name)
                
                    
            elif flag == 3:
                
                print(self.name, ' at ', today.strftime("%H:%M:%S"), ' : nighttime')
                self.quiet = 1
                self.DP = 0
                self.DA = 0
                this_receiver.nb_error = 0
                
            else : 
                
                print(self.name, ' is not transmitting at ', today.strftime("%H:%M:%S"))
                self.quiet = 1
        
                
        except Exception as e: # If there was an error, we update the log file
        
            this_receiver.nb_error += 1
            
            # Configuration of the logging file
            logging.basicConfig(filename='detection.log', encoding='utf-8')
            logging.error('Error at %s', 'division', exc_info=e)
            print(self.name, ' - error at ', today.strftime("%H:%M:%S"), ' - see log')
        
        if this_receiver.nb_error == 5: # it's been five minutes since the antenna stopped or had problems
            
            alerts_antenna.send_email()
        
    def get_pdf_coeff(self, path_to_probas):
        '''Returns the mu and sigma coefficients, defining a gaussian probabiility
        distribution function associated to a station and a value of DP
        
        :param path_to_probas: path to the files containing the mu and sigma values 
            for the porbability computations'''

        if self.DP < 0:
            self.DP = 0

        if self.DP > 5: # Only do it for high DP, the other ones may be due to small noise

            probas = pd.read_csv(path_to_probas + self.name + '_probas.csv')

            # Get the correct row
            rows = probas[probas['Low'] <= self.DP]
            row = rows.tail(1)
            self.mu = row['mu'].iloc[0]
            self.sigma = row['sigma'].iloc[0]

        else : # This is quiet time, we return the quiet proba centered in C1
 
            self.mu = -6 # log(1e-6)
            self.sigma = 1




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 14:28:49 2025

@author: pteysseyre
"""

import numpy as np


def analyse_breakpoint(current_time, current_p1, amp, phase,
                       the_time, quiet, quiet_phase, quiet_amp, last_time,
                       last_p1, detection_threshold, quiet_p1, quiet_breakpoint):

    """This is the part of the code that will decide if the  breakpoint is a
    flare signature or not.

    Several cases can occur:

        1) This is the beginning of a flare. 
        2) This is a flare maximum
        3) This is a quiet point: this is either a slight positive slope, a
         negative slope after a quiet breakpoint, or a negative
         slope after a flare has decayed by more than half of its increase
        4) This is a flare decay, with a negative slope and the
         phase has not gone down by more than half of its increase. 

    :param current_time: Time (in UT) of the last data reading
    :param current_p1: Slope of the new breakpoints
    :param amp: median amplitude measured over the last minute
    :param phase: phase data, since the beginning of the file or for the last two hours
    :param the_time: time-array, must match the one for phase
    :param quiet: status of the last breakpoint. If quiet==1, the last breakpoint was a quiet one, if not it was a flare time
    :param quiet_phase: phase at the last quiet breakpoint
    :param quiet_amp: amplitude at the last quiet breakpoint
    :param last_time: time of the last breakpoint OR two hours before if the last breakpoint was too long ago
    :param last_p1: Last breakpoint slope
    :param detection_threshold: attribute of the station class, slope above which we consider that it is flare time
    :param quiet_p1: slope of the last quiet breakpoint
    :param quiet_breakpoint: time of the last quiet breakpoint

    :return: - quiet_now (is 1 if the new breakpoint is considered to be quiet)
        - prev_timedecay which is the predicted time to decay (in UT)
        - DP_now (increase of phase (in °), compared to what the value of the phase should have been)
        - DA (value of the amplitude increase compared to last quiet amp)
    
        History:
        ----------------------
        - Written in February 2025
        - Revised in June 2025 to change the way DP was computed

        """
    # TODO: Think about the flares happening in a very fast sucession
    # Init: most of what is returned will not be systematically useful
    prev_timedecay = 0.0
    quiet_now = 0

    # Changed in June 2025 (to avoid having negatives DP)
    # DP_now = phase[-1] - quiet_phase 
    # Now the DP is computed using the last quiet breakpoint slope

    #TO CORRECT later if needed
    if quiet_p1 < 0:
        quiet_p1 = 0

    phase_nowifquiet = quiet_phase + quiet_p1*(current_time - quiet_breakpoint)
    DP_now = phase[-1] - phase_nowifquiet

    DA = amp - quiet_amp

    if current_p1 > 0:

        if (current_p1 < detection_threshold):

            if (quiet == 1):  # This is a quiet time
                quiet_now = 1
                DP_now = 0
                DA = 0

            else:  # That's either a flare max OR a flare that has decayed

                # We look for the phase maximum to get DP
                the_time, phase = _keep_between(phase, the_time, last_time,
                                                current_time)
                ind_max = np.argmax(phase)
                DP = phase[ind_max] - phase_nowifquiet
                time_max = the_time[ind_max]

                if DP_now < DP/2:  # The flare has already decayed
                    quiet_now = 1
                    DP_now = 0 

                else:  # This is the max of the flare
                    quiet_now = 0
                    prev_1min = DP
                    prev_5min = DP
        else:
            # This is a flare, but everything has already been done
            quiet_now = 0

    else:  # This is either quiet time, or a flare's decay

        if quiet:  # Last breakpoint is quiet
            quiet_now = True
            DP_now = 0

        else:  # Last breakpoint was a flare decay

            if last_p1 > 0:  # Last breakpoint was a flare onset

                # We look for the phase maximum to get DP
                the_time, phase = _keep_between(phase, the_time, last_time, 
                                                current_time)
                ind_max = np.argmax(phase)
                DP = phase[ind_max] - quiet_phase
                time_max = the_time[ind_max]

                # We estimate the decay time assuming that the trend is 
                # constant & the decay linear
                prev_timedecay = -DP/(2*current_p1) + time_max

            else:

                # Check if the flare has decayed 
                # (lost more than half its increase)
                ### TODO: change for Ne or abs

                # We look for the phase maximum to get DP
                the_time, phase = _keep_between(phase, the_time, last_time,
                                                 current_time)
                ind_max = np.argmax(phase)
                DP = phase[ind_max] - phase_nowifquiet
                time_max = the_time[ind_max]

                if DP_now < DP/2:  # The flare has ended
                    quiet_now = True
                    DP_now = 0

                else:  # The flare is ongoing but we can update our previsions
                    prev_timedecay = 1/current_p1*(DP/2 - DP_now) + the_time[-1]



    return (quiet_now, prev_timedecay, DP_now, DA)



def _keep_between(data, the_time, time_inf, time_sup):
    """ This is just a short bit of code to constrain the data between 
    time_inf & time_sup 
    
    :param data: data to be constrained
    :param the_time: time-array (the length must be the same as data)
    :param time_inf: Lowest limit to constrain the data array
    :param time_sup: Upper limit to constran the data array
    
    :return: the_time, new time-array constained between `time_inf` and `time_sup`
    :return: data, new data array constained between `time_inf` and `time_sup`"""

    if len(data) != len(the_time):

        print('Careful, lengths do not match - keep_between')

    else:

        ind_inf = np.argmin(np.abs(the_time - time_inf))
        ind_sup = np.argmin(np.abs(the_time - time_sup))

        the_time = the_time[ind_inf:ind_sup]
        data = data[ind_inf:ind_sup]

        return the_time, data















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 21 10:29:16 2024

@author: pteysseyre
"""

import numpy as np
import datetime as dt
from datetime import datetime, timedelta # To get today's date
import vlf4ions.read_files as rf

def correctionphase(station, phase, the_time):
    """Completely corrects the phase in real time the phase so that all phase jumps from -180° to 180° are corrected, and the phase detrended

    :param station: station class instance
    :param phase: phase data (in °)
    :param the_time: time-array in UT (or at least in hr)

    :return: phase_corr, corrected phase data
    """
    
    phase_corr, jumps = correctionphaseparDw(station, phase, the_time)
    
    return phase_corr


def unwrap_moco(Data_In, UnwrapNumber):
    """Originally written by M. Cohen in Matlab. Unwraps 'Data_in', 
    by compensating any jumps more than UnwrapNumer/2
    
    :param Data_In: Data to unwrap
    :param UnwrapNumber: Value of the jumps to compensate
    
    :return: Unwrapped data and position of the jumps"""
    
    # Remove all NaN point and reconcatanate
    Data_NaN = np.isnan(Data_In)
    Data_In_NonNaN = Data_In[Data_NaN == False]
    Indices_NonNaN = np.arange(0, np.size(Data_In_NonNaN))
    
    # Take differential
    Data_In_NonNaN_Diff = np.diff(Data_In_NonNaN)
    
    # Identify points where signal jumps by more than unwrap/2 threshold
    Indices_NonNaN = Indices_NonNaN[0:-1]
    JumpIndices = Indices_NonNaN[np.abs(Data_In_NonNaN_Diff)>UnwrapNumber/2]
    Data_In_NonNaN_Diff_Jumps = Data_In_NonNaN_Diff[JumpIndices]
    
    # Round jumps to nearest value of unwrap number
    Data_In_NonNaN_Diff_Jumps = np.round(Data_In_NonNaN_Diff_Jumps/UnwrapNumber) * UnwrapNumber
    
    # Integrates jump to original concatanated vector. Add 1 to indices to reference later point of two differential elements
    Indices_NonNaN_Jump = Indices_NonNaN[JumpIndices]
    
    # Go back to original version with the NaNs and create derivative
    Data_In_Diff_Jumps = Data_In*0
    Data_In_Diff_Jumps[np.isnan(Data_In_Diff_Jumps)] = 0
    Data_In_Diff_Jumps[Indices_NonNaN_Jump] = Data_In_NonNaN_Diff_Jumps
    
    # Integrates to get the function that needs to be substracted
    Data_In_Jumps = np.cumsum(Data_In_Diff_Jumps)
    
    # Substract from original 
    Data_Out = Data_In - Data_In_Jumps
    
    # nb_jumps_station = np.size(JumpIndices)
    
    return Data_Out, Data_In_Jumps[-1]

def correctionphaseparDw(station, phase, the_time):
    """Corrects the phase by compensating for the phase detrend induced by the 
    transmitter frequency not being exactly stable
    
    :param station: station class instance
    :param phase: phase data array (in °)
    :param the_time: time-array (in hr or UT)
    
    :return: phase_corr (unwrapped and detrended phase) and jumps (position of the 90° jumps)"""

    Deltaf = station.df
        
    Deltaw = 2*np.pi*Deltaf
    
    phase_un, jumps = unwrap_moco(phase, 90) 
    
    phase_corr = phase_un + (Deltaw*the_time*3600)*180/np.pi
    
    # For now (03/24) because there is both a 90° uncertainty on the measurements and the modelling
    # we choose to keep the phase between 90 and 180° for quiet daytime conditions
    
    
    return phase_corr, jumps

def compute_df(date_inf, date_sup, station, path_to_files, file_endings, epsilon):

    ''' Compute the value of df for a specific station over a specified time period

    :param date_inf: Datetime object, date from which we want to compute df
    :param date_sup: Datetime object, date before which we want to compute df
    :param station: Station class instance
    :param path_to_files: Path to the data files
    :param file_endings: List of string. Endings of the files for amp_NS, phase_NS, amp_EW and phase_EW in that order
    :param epsilon: Maximum phase jump (in °) that we accept in average from one day to the next

    '''

    # Compute number of days between date_inf and date_sup
    delta = date_sup - date_inf
    nb_days = delta.days

    # We check that there is enough data
    if nb_days < 2:
        print('Not enough days to compute df')

    # At the start, df is the one given in `station` 

    df_inf = -1e-2
    df_sup = 1e-2 # in Hz

    eps = epsilon*2 # Initialisation
    loop_counter = 0

    while eps>epsilon: # This is the bissection method

        loop_counter += 1

        eps = _loop_for_compute_df(nb_days, date_inf, station, path_to_files, file_endings)

        if np.abs(eps) < epsilon or loop_counter > 1e4 : 
            return station.df
        
        elif eps < 0: # df is too small

            df_inf = station.df
            station.df = (station.df + df_sup)/2
        
        else: # df is too large

            df_sup = station.df
            station.df = (station.df + df_inf)/2
    
    return station.df
        

def _loop_for_compute_df(nb_days, date_inf, station, path_to_files, file_endings):

    ''' This is just a short function to make compute_df easier to read
    
    :param nb_days: Number of days on which to loop
    :param date_inf: Datetime object, date from which we want to compute df
    :param station: Station class instance
    :param path_to_files: Path to the data files
    :param file_endings: List of string. Endings of the files for amp_NS, phase_NS, amp_EW and phase_EW in that order '''

    all_jumps = 0 # This will be the quantity to minimise
    last_phase_value = 0 # Last value of the phase on the last day

    days_with_data = nb_days

    for i in range(nb_days):

        # Day of interest, in the correct format
        today = date_inf + dt.timedelta(days=i)
        la_date = today.strftime("%Y_%m_%d")

        # Read the file 

        try:
            amp_NS, amp_EW, phase_NS, phase_EW, the_time = rf.read_Narrowband_postprocessing(la_date, station.name, path_to_files, file_endings)


            if station.orientation==1:
                phase = phase_EW
            else:
                phase = phase_NS
            phase, jumps = correctionphaseparDw(station, phase, the_time)

            if i > 0: # We do not make this calculation for the first day

                all_jumps += (phase[-1] - last_phase_value)
            
            last_phase_value = phase[-1] # We keep track of the last value of the day
        
        except : # There is no data for this day

            print('Data not available on ' + la_date)
            last_phase_value = np.nan
            days_with_data = days_with_data - 1
    
    all_jumps = all_jumps/(days_with_data - 1) # There are nb_days - 1 midnight crossings 

    return all_jumps
        

